/**********************************************************************/
/*                                                                    */
/* HTML2TXT -- filter to convert (ISO 8859-1) HTML to PC-850 text.    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* A stand-alone implementation of the RexxMail HTMLText procedure.   */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* The object of the procedure is to obtain a useful PC850 text       */
/* rendition of the readable content of an HTML file.                 */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Usage : htmltext < infile [> outfile] | [| program]                */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* This program extracts text content from an HTML file as follows:   */
/*                                                                    */
/* - Input is expected to be ISO-8859-1 HTML;                         */
/* - HTML text codes (&[name];) are converted to PC850 equivalents;   */
/* - HTML numeric codes (&#nnn;) are converted to PC850 equivalents;  */
/* - The special numeric euro code (#8364) is supported;              */
/* - Formatting of text between <PRE> and </PRE> tags is preserved;   */
/* - Tables are converted to straight text, with marker lines before  */
/*   and after data sections;                                         */
/* - Significant tags (e.g. HR, Hn, TD) are shown as marker lines;    */
/* - Lists (definition, ordered, and unordered) are supported.        */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* This program is released under the terms of the GNU license, see   */
/*               www.gnu.org/copyleft/gpl.html                        */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* (c) 2000-2002 Marcus de Geus                                       */
/*               marcus@degeus.com                                    */
/*               www.degeus.com                                       */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Use it if you like it. Don't if you don't. No legalese.            */
/*                                                                    */
/**********************************************************************/


/**********************************************************************/
/* Some definitions (the "Global." is to make life easier for me)     */
/**********************************************************************/

Global.ISOto850 = '_______________'||,  /* bytes 80 to 8F */
                  '"&<>____________'||,  /* bytes 90 to 9F */
                  'Ͼ'||,  /* bytes A0 to AF */
                  ''||,  /* bytes B0 to BF */
                  'ǎԐ'||,  /* bytes C0 to CF */
                  'ѥ噞'||,  /* bytes D0 to DF */
                  'Ƅ'||,  /* bytes E0 to EF */
                  'Ф'  /* bytes F0 to FF */

Global.CRLF = d2c(13)||d2c(10)  /* define CRLF */

/**********************************************************************/
/* Main                                                               */
/**********************************************************************/

parse arg Global.ASCIILineLength  /* look for a parameter */

if (Global.ASCIILineLength = '') then  /* if there is none */
do
 Global.ASCIILineLength = 78  /* the default is 78 */
end
else  /* if we have something */
do
 if (\datatype(Global.ASCIILineLength,'W')) then  /* if it is not a whole number */
 do
  call beep 333,333  /* signal */
  say 'Usage : HTML2TXT [linelength] < inputfile [[> outputfile] | [| program]]'  /* report */
  call Halt  /* and quit */
 end
end

signal on Halt  /* handle [Ctrl][C] */

HTMLText = ''  /* start with nothing */

do while lines('STDIN:')  /* as long as we have lines coming in */
 HTMLText = HTMLText||linein('STDIN:')||Global.CRLF  /* get the HTML content from standard input */
end

call charout 'STDOUT:',HTMLText(HTMLText)  /* call the HTMLText procedure to do the conversion and write the results to standard output */

call Halt  /* that's all, folks! */

/**********************************************************************/
Halt:  /* handles halt condition */
/**********************************************************************/

exit 0  /* just quit */

/**********************************************************************/
WordWrap: procedure expose Global.  /* word-wraps a line of text */
/**********************************************************************/

parse arg TextLine,MaxLength,Indent,NoBreak  /* get the arguments */

NoBreak = (NoBreak = 1)  /* this one is true if the argument is 1 */

if (Indent = '') then  /* if we have no indent */
do
 Indent = 0  /* we have no indent */
end

if (MaxLength = '') then  /* if we have no maximum line length value */
do
 MaxLength = 76  /* use this */
end

CutPoint = MaxLength + 1  /* start the cut at ASCIILineLength plus one */

NewText = ''  /* start with nothing */

TextLine = strip(TextLine,'T',' ')  /* remove any trailing spaces */

do while (length(TextLine) > MaxLength)  /* as long as the line is longer than the limit */

 parse var TextLine NextBit =(CutPoint) TextLine  /* get the next bit of the maximum length */

 if ((substr(NextBit,length(NextBit),1) = ' ') | (substr(TextLine,1,1) = ' ')) then  /* if we cut the line just before or after a space, i.e. we cut between two words */
 do
  TextLine = strip(TextLine,'L',' ')  /* get rid of leading spaces on the remaining bit of the line */
 end
 else  /* if we cut through a word */
 do
  LastSpace = lastpos(' ',NextBit)  /* look for the last space in this bit */

  if (LastSpace >< 0) then  /* if we found a space */
  do
   parse var NextBit NextBit =(LastSpace) MoreBit  /* cut off the bit after the space and save it for the next line */
   TextLine = strip(MoreBit,'L',' ')||TextLine  /* get rid of the leading space and stick this bit back on to what's left of the original text line */
  end
  else  /* if we found no space */
  do
   if (NoBreak) then  /* if we must not break words */
   do
    parse var TextLine MoreBit TextLine  /* get the first word of the remaining bit of the line */
    NextBit = NextBit||MoreBit  /* and add it to what we have */
   end
  end

 end

 NextBit = strip(NextBit,'T',' ')  /* get rid of trailing spaces */
 NewText = NewText||NextBit  /* add the bit we have to the new text */

 if (TextLine >< '') then  /* if there is more to come */
 do
  NewText = NewText||Global.CRLF||copies(' ',Indent)  /* add a new line with an indent */
 end
end

NewText = NewText||TextLine  /* add the remaining bit of the line to the text */

return NewText  /* end of WordWrap */

/**********************************************************************/
HTMLText: procedure expose Global.  /* converts (ISO 8859-1) HTML to PC-850 text */
/**********************************************************************/

parse arg HTMLText  /* get the argument */

HTMLCodes = 'euro',  /* euro sign */
            copies('_ ',15),  /* 15 blanks */
            'quot',  /* double quotation mark */
            'amp',  /* ampersand */
            'lt',  /* less-than sign */
            'gt',  /* greater-than sign */
            copies('_ ',12),  /* 12 blanks */
            'nbsp',  /* non-breaking space */
            'iexcl',  /* inverted exclamation mark */
            'cent',  /* cent sign */
            'pound',  /* pound sterling sign */
            'curren',  /* general currency sign */
            'yen',  /* yen sign */
            'brvbar',  /* broken vertical bar */
            'sect',  /* section sign */
            'uml',  /* umlaut (dieresis) */
            'copy',  /* copyright */
            'ordf',  /* feminine ordinal */
            'laquo',  /* left angle quote, guillemot left */
            'not',  /* not sign */
            'shy',  /* soft hyphen */
            'reg',  /* registered trademark */
            'macr',  /* macron accent */
            'deg',  /* degree sign */
            'plusmn',  /* plus or minus */
            'sup2',  /* superscript two */
            'sup3',  /* superscript three */
            'acute',  /* acute accent */
            'micro',  /* micro sign */
            'para',  /* paragraph sign */
            'middot',  /* middle dot */
            'cedil',  /* cedilla */
            'sup1',  /* superscript one */
            'ordm',  /* masculine ordinal */
            'raquo',  /* right angle quote, guillemot right */
            'frac14',  /* fraction one quarter */
            'frac12',  /* fraction one half */
            'frac34',  /* fraction three quarters */
            'iquest',  /* inverted question mark */
            'Agrave',  /* uppercase A, grave */
            'Aacute',  /* uppercase A, acute */
            'Acirc',  /* uppercase A, circumflex */
            'Atilde',  /* uppercase A, tilde */
            'Auml',  /* uppercase A, dieresis or Umlaut */
            'Aring',  /* uppercase A, ring */
            'AElig',  /* uppercase AE diphthong (ligature) */
            'Ccedil',  /* uppercase C, cedilla */
            'Egrave',  /* uppercase E, grave */
            'Eacute',  /* uppercase E, acute */
            'Ecirc',  /* uppercase E, circumflex */
            'Euml',  /* uppercase E, dieresis or Umlaut */
            'Igrave',  /* uppercase I, grave */
            'Iacute',  /* uppercase I, acute */
            'Icirc',  /* uppercase I, circumflex */
            'Iuml',  /* uppercase I, dieresis or Umlaut */
            'ETH',  /* uppercase Eth, Icelandic */
            'Ntilde',  /* uppercase N, tilde */
            'Ograve',  /* uppercase O, grave */
            'Oacute',  /* uppercase O, acute */
            'Ocirc',  /* uppercase O, circumflex */
            'Otilde',  /* uppercase O, tilde */
            'Ouml',  /* uppercase O, dieresis or Umlaut */
            'times',  /* multiply sign */
            'Oslash',  /* uppercase O, slash */
            'Ugrave',  /* uppercase U, grave */
            'Uacute',  /* uppercase U, acute */
            'Ucirc',  /* uppercase U, circumflex */
            'Uuml',  /* uppercase U, dieresis or Umlaut */
            'Yacute',  /* uppercase Y, acute */
            'THORN',  /* uppercase THORN, Icelandic */
            'szlig',  /* lowercase sharp s, sz ligature */
            'agrave',  /* lowercase a, grave */
            'aacute',  /* lowercase a, acute */
            'acirc',  /* lowercase a, circumflex */
            'atilde',  /* lowercase a, tilde */
            'auml',  /* lowercase a, dieresis or Umlaut */
            'aring',  /* lowercase a, ring */
            'aelig',  /* lowercase ae diphthong (ligature) */
            'ccedil',  /* lowercase c, cedilla */
            'egrave',  /* lowercase e, grave */
            'eacute',  /* lowercase e, acute */
            'ecirc',  /* lowercase e, circumflex */
            'euml',  /* lowercase e, dieresis or Umlaut */
            'igrave',  /* lowercase i, grave */
            'iacute',  /* lowercase i, acute */
            'icirc',  /* lowercase i, circumflex */
            'iuml',  /* lowercase i, dieresis or Umlaut */
            'eth',  /* lowercase eth, Icelandic */
            'ntilde',  /* lowercase n, tilde */
            'ograve',  /* lowercase o, grave */
            'oacute',  /* lowercase o, acute */
            'ocirc',  /* lowercase o, circumflex */
            'otilde',  /* lowercase o, tilde */
            'ouml',  /* lowercase o, dieresis or Umlaut */
            'divide',  /* division sign */
            'oslash',  /* lowercase o, slash */
            'ugrave',  /* lowercase u, grave */
            'uacute',  /* lowercase u, acute */
            'ucirc',  /* lowercase u, circumflex */
            'uuml',  /* lowercase u, dieresis or Umlaut */
            'yacute',  /* lowercase y, acute */
            'thorn',  /* lowercase thorn, Icelandic */
            'yuml'  /* lowercase y, dieresis or umlaut */

IndentTags = 'DL',  /* <DL> */
             'OL',  /* <OL> */
             'UL'  /* <UL> */

UnIndentTags = '/DL',  /* </DL> */
               '/OL',  /* </OL> */
               '/UL'  /* </UL> */

NewLineTags = 'BR',  /* <BR> */
              'DIV',  /* <DIV> */
              'P',  /* <P> */
              'HR',  /* <HR> */
              'DT',  /* <DT> */
              'DD',  /* <DD> */
              'LI',  /* <LI> */
              'H1',  /* <H1> */
              '/H1',  /* </H1> */
              'H2',  /* <H2> */
              '/H2',  /* </H2> */
              'H3',  /* <H3> */
              '/H3',  /* </H3> */
              'H4',  /* <H4> */
              '/H4',  /* </H4> */
              'H5',  /* <H5> */
              '/H5',  /* </H5> */
              'H6',  /* <H6> */
              '/H6',  /* </H6> */
              'TD',  /* <TD> */
              '/TD',  /* </TD> */
              'PRE',  /* <PRE> */
              '/PRE',  /* </PRE> */
              'BLOCKQUOTE',  /* <BLOCKQUOTE> */
              '/BLOCKQUOTE',  /* </BLOCKQUOTE> */
              'TITLE',  /* <TITLE> */
              '/TITLE',  /* </TITLE> */
              '/HTML'  /* </HTML> */

Text = ''  /* start with nothing */
AddText = ''  /* start with nothing to add */
TypeStack = ''  /* start with an empty type stack */
PreText = 0  /* no preformatted text yet */
Indent = 0  /* start with no indent */

do while (HTMLText >< '')  /* as long as we have HTML text left */

 parse var HTMLText NextPart '<' TagCont '>' HTMLText  /* get the next part before any tag, and keep the tag content */

 if (NextPart >< '') then  /* if we have something */
 do

   do while (NextPart >< '')  /* as long as we have something left */

   parse var NextPart BeforePart '&' CharCode ';' NextPart  /* get the bit before any character code, and keep the code */

   if (CharCode >< '') then  /* if we have a character code */
   do

    if ((substr(CharCode,1,1) = '#') & (datatype(substr(CharCode,2),'W'))) then  /* if the code is a character number */
    do

     if (substr(CharCode,2) = '8364') then  /* if it is the euro code */
     do
      Character = d2c(213)  /* use the euro symbol */
     end
     else  /* if it is not the euro code */
     do
      Character = translate(d2c(substr(CharCode,2)),Global.ISOto850,xrange('80'x,'FF'x))  /* convert the ISO-8859-1 code to the PC-850 equivalent */
     end

    end
    else  /* if the character code is not a number */
    do

     Position = wordpos(CharCode,HTMLCodes)  /* look for the character code in the HTML codes string */

     if (Position = 0) then  /* if we do not find it */
     do
      Character = '_'  /* use this */
     end
     else  /* if we do find it */
     do
      Character = substr(Global.ISOto850,Position,1)  /* get the character we want from the ISOto850 array */
     end

    end

    BeforePart = BeforePart||Character  /* add the character to what we have */

   end

   AddText = AddText||BeforePart  /* add the result to what we have */

  end

  if (TagCont = '') then  /* if we have no tag following the text */
  do
   TagCont = '/HTML'  /* force out any remaining text */
  end

 end

 if (TagCont >< '') then  /* if we have tag content */
 do

  TagWord = translate(word(TagCont,1))  /* get the first word of the tag content in upper case */

  if (PreText) then  /* if we are dealing with preformatted text */
  do

   select  /* do one of the following */

    when (TagWord = 'A') then  /* if the tag is an anchor, it is a ligitimate tag and can be dropped */
    do
     nop  /* ignore it */
    end

    when (TagWord = '!--') then  /* if the tag is a comment, it is a ligitimate tag and can be dropped */
    do
     nop  /* ignore it */
    end

    when (TagWord = '/PRE') then  /* if the tag signals the end of preformatted text */
    do
     do while (AddText >< '')  /* walk through the part we have collected */
      parse var AddText NextBit (Global.CRLF) AddText  /* get the bit up to the next CRLF */
      Text = Text||copies(' ',(Indent * 4))||NextBit||Global.CRLF  /* add it to what we have, with an indent */
     end
     PreText = 0  /* reset the preformatted text flag */
    end

    otherwise  /* if it is not a legitimate tag, assume it is part of the text, which may contain a real tag later on */
    do
     AddText = AddText||'<'  /* add the left angle bracket to what we have */
     HTMLText = TagCont||'>'||HTMLText  /* restore the remainder to be processed next time around */
    end

   end

  end
  else  /* if we are not dealing with preformatted text */
  do

   if (wordpos(TagWord,IndentTags||' '||UnIndentTags||' '||NewLineTags) > 0) then  /* if it is one of these */
   do

    NewText = ''  /* start with nothing */

    if (AddText >< '') then  /* if we have something to add */
    do
     AddText = strip(translate(AddText,'   ',d2c(9)||Global.CRLF),'B',' ')  /* take what we have, turn TAB and CRLF into space, and remove excess space */
     do while (AddText >< '')  /* as long as we have something left */
      parse var AddText NextWord AddText  /* get the next word (to get rid of excess whitespace) */
      NewText = NewText||NextWord||' '  /* and add it to what we have, followed by a space */
     end
     NewText = strip(NewText,'T',' ')  /* remove the trailing blank */
    end

    if (NewText >< '') then  /* if we have something to add */
    do

     if (Text >< '') then  /* if previous text exists */
     do
      if (substr(Text,(length(Text) - 1),2) = Global.CRLF) then  /* if the previous text ends with a CRLF */
      do
       Text = Text||copies(' ',(Indent * 4))  /* add a full indent */
      end
     end

     Text = Text||WordWrap(NewText,(Global.ASCIILineLength - (Indent * 4)),(Indent * 4))||Global.CRLF  /* word-wrap the new text part with a full indent and add it to the existing text, followed by a new line */

    end
    else  /* if we have nothing to add */
    do

     if ((wordpos(TagWord,NewLineTags) > 0) & (Text >< '')) then  /* if it is one of these and we already have text */
     do

      if (substr(Text,(length(Text) - 1),2) >< Global.CRLF) then  /* if the previous text does not end with a CRLF */
      do
       Text = Text||Global.CRLF  /* add a new line */
      end

     end

    end

    select  /* do one of the following */

     when (wordpos(TagWord,IndentTags) > 0) then  /* if it is one of these */
     do
      TypeStack = TagWord||' '||TypeStack  /* push the tag word onto the type stack */
      Indent = Indent + 1  /* move up */
      if (TagWord = 'OL') then  /* if we are starting an ordered list */
      do
       OListCount.Indent = 0  /* start a new list counter */
      end
     end

     when (wordpos(TagWord,UnIndentTags) > 0) then  /* if it is one of these */
     do
      if (TagWord = '/OL') then  /* if we are ending an ordered list */
      do
       drop OListCount.Indent  /* drop the list counter */
      end
      Indent = Indent - 1  /* move back */
      TypeStack = subword(TypeStack,2)  /* remove the tag word from the type stack */
     end

     otherwise  /* if must be a new line tag */
     do

      select  /* do one of the following */

       when (TagWord = 'P') then  /* if it is a paragraph marker */
       do
        Text = Text||Global.CRLF  /* add another CRLF */
       end

       when (TagWord = 'PRE') then  /* if it is the start of preformatted text */
       do
        PreText = 1  /* set the preformatted text flag */
       end

       when (TagWord = 'DT') then  /* if it is a definition tag */
       do
        Text = Text||copies(' ',((Indent - 1) * 4))||'  - '  /* indent minus one plus a marker */
       end

       when (TagWord = 'DD') then  /* if it is a definition definition */
       do
        Text = Text||copies(' ',(Indent * 4))  /* full indent */
       end

       when (TagWord = 'LI') then  /* if it is a list item */
       do

        if (word(TypeStack,1) = 'OL') then  /* if we are in an ordered list */
        do
         OListCount.Indent = OListCount.Indent + 1  /* up the current list counter */
         Text = Text||copies(' ',((Indent - 1) * 4))||right(OListCount.Indent,2,' ')||'. '  /* indent minus one, the number, a full stop, and a space */
        end
        else  /* we must be in an unordered list */
        do
         Text = Text||copies(' ',((Indent - 1) * 4))||'   '  /* indent minus one, and a bullet */
        end

       end

       when (wordpos(TagWord,'HR H1 H2 H3 H4 H5 H6 TD TITLE') > 0) then  /* if it is one of these */
       do

        Text = Text||copies(' ',(Indent * 4))||copies('_',(Global.ASCIILineLength - (Indent * 4)))||Global.CRLF  /* full indent, a starting line and a new line */

        if (TagWord = 'TITLE') then  /* if it is a document title */
        do
         Text = Text||copies(' ',(Indent * 4))||'Document title : '  /* add an indent and some text */
        end
        else  /* if it is not a document title */
        do
         if (TagWord = 'HR') then  /* if it is a separator line we want */
         do
          Text = Text||copies(' ',(Indent * 4))||copies('',(Global.ASCIILineLength - (Indent * 4)))||Global.CRLF  /* add an ending line */
         end
        end

       end

       when (wordpos(TagWord,'/H1 /H2 /H3 /H4 /H5 /TD /TITLE') > 0) then  /* if it is one of these */
       do
        Text = Text||copies(' ',(Indent * 4))||copies('',(Global.ASCIILineLength - (Indent * 4)))||Global.CRLF  /* add an ending line */
       end

       otherwise  /* if it is none of the above */
       do
        nop  /* do nothing else */
       end

      end

     end

    end

   end

  end

 end

end

return Text  /* end of HTMLText */
